/* $Id: cpl_mask-test.c,v 1.31 2008/12/02 14:02:48 llundin Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2008 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2008/12/02 14:02:48 $
 * $Revision: 1.31 $
 * $Name: cpl-5_0_0 $
 */

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <math.h>

#include "cpl_mask.h"
#include "cpl_image_gen.h"
#include "cpl_stats.h"
#include "cpl_image_io.h"
#include "cpl_memory.h"
#include "cpl_msg.h"
#include "cpl_test.h"

/*-----------------------------------------------------------------------------
                                  Defines
 -----------------------------------------------------------------------------*/

#ifndef IMAGESZ
#define IMAGESZ 512
#endif

#ifndef MASKSZ
#define MASKSZ 8
#endif

/*-----------------------------------------------------------------------------
                                  Private functions
 -----------------------------------------------------------------------------*/

static cpl_error_code cpl_mask_shift_bf(cpl_mask *, int, int);
static void cpl_mask_shift_test(int, int);

static double tshift   = 0.0;
static double tshiftbf = 0.0;

/*-----------------------------------------------------------------------------
                                  Main
 -----------------------------------------------------------------------------*/
int main(void)
{
    cpl_image       *   imd;
    double              median,
                        med_dist;
    cpl_mask        *   bin1,
                    *   bin2,
                    *   bin1_dup;
    cpl_image       *   labels1,
                    *   labels2;
    double              threshold;
    int                 count;
    int                 nb_objs1,
                        nb_objs2;
    cpl_matrix      *   kernel;
    double          *   ker_arr;
    int                 new_pos[] = {1,2,3,4};
    /* char                outname[1024]; */
    const int           nsize = IMAGESZ;
    cpl_binary      *   bins;
    cpl_error_code      error;

    cpl_test_init(PACKAGE_BUGREPORT, CPL_MSG_WARNING);

    /* Insert tests below */

    cpl_mask_shift_test(1, 1);
    cpl_mask_shift_test(1, 2);
    cpl_mask_shift_test(2, 1);
    cpl_mask_shift_test(2, 2);
    cpl_mask_shift_test(MASKSZ, 1);
    cpl_mask_shift_test(1, MASKSZ);
    cpl_mask_shift_test(MASKSZ, 2);
    cpl_mask_shift_test(2, MASKSZ);
    cpl_mask_shift_test(MASKSZ, 3);
    cpl_mask_shift_test(3, MASKSZ);
    cpl_mask_shift_test(MASKSZ, MASKSZ);

    if (cpl_msg_get_level() < CPL_MSG_WARNING) {
        cpl_mask_shift_test(IMAGESZ, MASKSZ);
    }

    cpl_msg_info(cpl_func, "Time to shift [s]: %g <=> %g (brute force)",
                 tshift, tshiftbf);

    /* Create test image */
    cpl_msg_info("","Create a %dx%d test image", nsize, nsize);
    imd = cpl_image_fill_test_create(nsize, nsize);
    cpl_test_nonnull(imd);
    
    /* Compute the median and the av. dist. to the median of the loaded image */
    median = cpl_image_get_median_dev(imd, &med_dist);

    /* Binarize the image with a first threshold */
    cpl_msg_info("","Threshold the test image with median+2*sigma -> bin1");
    threshold = median + 2.0 * med_dist;
    bin1 = cpl_mask_threshold_image_create(imd, threshold, DBL_MAX);
    cpl_test_nonnull(bin1);

    cpl_test_zero(cpl_mask_is_empty(bin1));

    /* Test wrap / unwrap */
    bins = cpl_calloc(100, sizeof(cpl_binary));
    bin2 = cpl_mask_wrap(10, 10, bins);
    cpl_test(cpl_mask_is_empty(bin2));

    cpl_mask_delete(bin2);
    bins = cpl_calloc(100, sizeof(cpl_binary));
    bin2 = cpl_mask_wrap(10, 10, bins);
    cpl_test(cpl_mask_is_empty(bin2));
    cpl_test(bins == cpl_mask_unwrap(bin2));
    cpl_free(bins);

    /* Binarize the image with a second threshold */
    cpl_msg_info("","Threshold the test image with median+5*sigma -> bin2");
    threshold = median + 5.0 * med_dist; 
    bin2 = cpl_mask_threshold_image_create(imd, threshold, DBL_MAX);
    cpl_test_nonnull(bin2);
    cpl_image_delete(imd);
   
    /* Test cpl_mask_get() */
    /* Test cpl_mask_set() */
    count = cpl_mask_count(bin1);
    if (cpl_mask_get(bin1, 1, 1) == CPL_BINARY_0) {
        cpl_test_zero(cpl_mask_count_window(bin1, 1, 1, 1, 1));
        cpl_mask_set(bin1, 1, 1, CPL_BINARY_1);
        cpl_test_eq(cpl_mask_count(bin1), count + 1);
        cpl_test_eq(cpl_mask_count_window(bin1, 1, 1, 1, 1), 1);
    } else {
        cpl_test_eq(cpl_mask_count_window(bin1, 1, 1, 1, 1), 1);
        cpl_mask_set(bin1, 1, 1, CPL_BINARY_0);
        cpl_test_eq(cpl_mask_count(bin1), count - 1);
        cpl_test_zero(cpl_mask_count_window(bin1, 1, 1, 1, 1));
        /* Later a non-zero mask is required */
        cpl_mask_set(bin1, 1, 1, CPL_BINARY_1);
    }

    /* Test error handling of cpl_mask_set() */
    error = cpl_mask_set(NULL, 1, 1, CPL_BINARY_1);
    cpl_test_error(error);
    cpl_test_eq(error, CPL_ERROR_NULL_INPUT);

    error = cpl_mask_set(bin1, 0, 1, CPL_BINARY_1);
    cpl_test_error(error);
    cpl_test_eq(error, CPL_ERROR_ILLEGAL_INPUT);

    error = cpl_mask_set(bin1, 1, 0, CPL_BINARY_1);
    cpl_test_error(error);
    cpl_test_eq(error, CPL_ERROR_ILLEGAL_INPUT);

    error = cpl_mask_set(bin1, 1, 1+nsize, CPL_BINARY_1);
    cpl_test_error(error);
    cpl_test_eq(error, CPL_ERROR_ILLEGAL_INPUT);

    error = cpl_mask_set(bin1, 1+nsize, 1, CPL_BINARY_1);
    cpl_test_error(error);
    cpl_test_eq(error, CPL_ERROR_ILLEGAL_INPUT);

    error = cpl_mask_set(bin1, 1, 1, (cpl_binary)-1);
    cpl_test_error(error);
    cpl_test_eq(error, CPL_ERROR_ILLEGAL_INPUT);

    /* Compute number of selected pixels */
    cpl_test_eq(cpl_mask_count(bin1),
                cpl_mask_count_window(bin1, 1, 1, nsize, nsize));
    
    /* Test cpl_mask_duplicate() */
    bin1_dup = cpl_mask_duplicate(bin1);
    cpl_test_eq_mask(bin1, bin1_dup);
    cpl_test_eq(cpl_mask_is_empty(bin1), cpl_mask_is_empty(bin1_dup));
    cpl_test_eq(cpl_mask_count(bin1),    cpl_mask_count(bin1_dup));

    /* Test cpl_mask_turn() */
    /* Rotate the mask all the way around */
    cpl_test(cpl_mask_turn(bin1,   9) == CPL_ERROR_NONE);
    cpl_test(cpl_mask_turn(bin1,   6) == CPL_ERROR_NONE);
    cpl_test(cpl_mask_turn(bin1, -13) == CPL_ERROR_NONE);
    cpl_test(cpl_mask_turn(bin1,  10) == CPL_ERROR_NONE);
    cpl_test(cpl_mask_turn(bin1,   4) == CPL_ERROR_NONE);
    cpl_test_eq_mask(bin1, bin1_dup);
    /* Compare to its copy */
    cpl_test_zero(cpl_mask_xor(bin1_dup, bin1));
    cpl_test_error(CPL_ERROR_NONE);
    cpl_test_zero(cpl_mask_count(bin1_dup));
    cpl_test_eq(cpl_mask_is_empty(bin1_dup), CPL_TRUE);
    cpl_mask_delete(bin1_dup);

    /* Build a 3x3 kernel */
    ker_arr = cpl_malloc(9*sizeof(double));
    ker_arr[0] = 1;
    ker_arr[1] = 1;
    ker_arr[2] = 1;
    ker_arr[3] = 1;
    ker_arr[4] = 1;
    ker_arr[5] = 1;
    ker_arr[6] = 1;
    ker_arr[7] = 1;
    ker_arr[8] = 1;
    kernel = cpl_matrix_wrap(3, 3, ker_arr);

    cpl_msg_info("","Apply a morphological closing on bin1");
    /* Test erosion function */
    cpl_test_zero(cpl_mask_erosion(bin1, kernel));
    cpl_test_error(CPL_ERROR_NONE);

    /* Test dilation function */
    cpl_test_zero(cpl_mask_dilation(bin1, kernel));
    cpl_test_error(CPL_ERROR_NONE);

    cpl_msg_info("","Apply a morphological opening on bin1");
    /* Test opening function */
    cpl_test_zero(cpl_mask_opening(bin1, kernel));
    cpl_test_error(CPL_ERROR_NONE);
 
    cpl_msg_info("","Apply a morphological closing on bin2");
    /* Test closing function */
    cpl_test_zero(cpl_mask_closing(bin1, kernel));
    cpl_test_error(CPL_ERROR_NONE);

    cpl_msg_info("","Apply a morphological opening on bin2");
    /* Test dilation function */
    cpl_test_zero(cpl_mask_dilation(bin2, kernel));
    cpl_test_error(CPL_ERROR_NONE);


    /* Test erosion function */
    cpl_test_zero(cpl_mask_erosion(bin2, kernel));
    cpl_test_error(CPL_ERROR_NONE);

    cpl_matrix_delete(kernel);

    /* Labelize the binary maps */
    labels1 = cpl_image_labelise_mask_create(bin1, &nb_objs1);
    cpl_test_nonnull(labels1);

    cpl_msg_info("","Labelize bin1 ---> %d objects found", nb_objs1);
    labels2 = cpl_image_labelise_mask_create(bin2, &nb_objs2);
    cpl_test_nonnull(labels2);

    cpl_msg_info("","Labelize bin2 ---> %d objects found", nb_objs2);
   
    /* sprintf(outname, "labelized1.fits"); */
    /* cpl_image_save(labels1, outname, CPL_BPP_IEEE_FLOAT, NULL,
                      CPL_IO_CREATE); */
    /* sprintf(outname, "labelized2.fits"); */
    /* cpl_image_save(labels2, outname, CPL_BPP_IEEE_FLOAT, NULL,
                      CPL_IO_CREATE); */
    
    cpl_image_delete(labels1);
    cpl_image_delete(labels2);

    /* The following tests need to be done on a non-empty mask */ 
    cpl_test_zero(cpl_mask_is_empty(bin2));

    /* Test cpl_mask_and */
    cpl_test_zero(cpl_mask_and(bin1, bin2));
    cpl_test_error(CPL_ERROR_NONE);
    cpl_test_leq(cpl_mask_count(bin1), cpl_mask_count(bin2));
    
    /* Test cpl_mask_or */
    cpl_test_zero(cpl_mask_or(bin1, bin2));
    cpl_test_error(CPL_ERROR_NONE);

    cpl_test_eq_mask(bin1, bin2);
     
    /* Test cpl_mask_not */
    cpl_test_zero(cpl_mask_not(bin2));
    cpl_test_error(CPL_ERROR_NONE);

    /* Test cpl_mask_xor */
    cpl_test_zero(cpl_mask_xor(bin1, bin2));
    cpl_test_error(CPL_ERROR_NONE);

    /* bin1 is now all ones... */

    /* Test cpl_mask_copy() */
    cpl_test_zero(cpl_mask_copy(bin2, bin1, 1, 1));
    cpl_test_error(CPL_ERROR_NONE);

    cpl_test_eq_mask(bin1, bin2);

    cpl_test_zero(cpl_mask_not(bin2));
    cpl_test_error(CPL_ERROR_NONE);

    cpl_test(cpl_mask_is_empty(bin2));

    /* Test error handling in cpl_mask_copy() */
    error = cpl_mask_copy(NULL, bin2, 1, 1);
    cpl_test_error(CPL_ERROR_NULL_INPUT);
    cpl_test_eq(error, CPL_ERROR_NULL_INPUT);

    error = cpl_mask_copy(bin1, NULL, 1, 1);
    cpl_test_error(CPL_ERROR_NULL_INPUT);
    cpl_test_eq(error, CPL_ERROR_NULL_INPUT);

    error = cpl_mask_copy(bin2, bin1, 0, 1);
    cpl_test_error(CPL_ERROR_ILLEGAL_INPUT);
    cpl_test_eq(error, CPL_ERROR_ILLEGAL_INPUT);

    error = cpl_mask_copy(bin2, bin1, 1, 0);
    cpl_test_error(CPL_ERROR_ILLEGAL_INPUT);
    cpl_test_eq(error, CPL_ERROR_ILLEGAL_INPUT);

    error = cpl_mask_copy(bin2, bin1, nsize+1, 1);
    cpl_test_error(CPL_ERROR_ILLEGAL_INPUT);
    cpl_test_eq(error, CPL_ERROR_ILLEGAL_INPUT);

    error = cpl_mask_copy(bin2, bin1, 1, nsize+1);
    cpl_test_error(CPL_ERROR_ILLEGAL_INPUT);
    cpl_test_eq(error, CPL_ERROR_ILLEGAL_INPUT);

    cpl_mask_delete(bin2);
    
    /* Test cpl_mask_collapse_create() */
    bin2 = cpl_mask_collapse_create(bin1, 0);
    cpl_test_nonnull(bin2);
    cpl_test_eq(cpl_mask_get_size_x(bin2), nsize);
    cpl_test_eq(cpl_mask_get_size_y(bin2), 1);

    bin1_dup = cpl_mask_collapse_create(bin2, 0);
    cpl_test_nonnull(bin1_dup);
    cpl_test_eq(cpl_mask_get_size_x(bin1_dup), nsize);
    cpl_test_eq(cpl_mask_get_size_y(bin1_dup), 1);
    cpl_mask_delete(bin1_dup);

    bin1_dup = cpl_mask_collapse_create(bin2, 1);
    cpl_test_nonnull(bin1_dup);
    cpl_test_eq(cpl_mask_get_size_x(bin1_dup), 1);
    cpl_test_eq(cpl_mask_get_size_y(bin1_dup), 1);
    cpl_mask_delete(bin1_dup);
    cpl_mask_delete(bin2);
    
    /* Test cpl_mask_extract() */
    bin2 = cpl_mask_extract(bin1, 1, 1, 1, 1);
    cpl_test_nonnull(bin2);
    cpl_test_eq(cpl_mask_get_size_x(bin2), 1);
    cpl_test_eq(cpl_mask_get_size_y(bin2), 1);
    cpl_mask_delete(bin2);
    
    /* Test cpl_mask_shift() */
    cpl_test_zero(cpl_mask_shift(bin1, 1, 1));
    cpl_test_error(CPL_ERROR_NONE);
    
    bin2 = cpl_mask_duplicate(bin1);

    /* Test cpl_mask_flip() */
    cpl_test_zero(cpl_mask_flip(bin1, 0));
    cpl_test_error(CPL_ERROR_NONE);
    cpl_test_zero(cpl_mask_flip(bin1, 1));
    cpl_test_error(CPL_ERROR_NONE);
    cpl_test_zero(cpl_mask_flip(bin1, 2));
    cpl_test_error(CPL_ERROR_NONE);
    cpl_test_zero(cpl_mask_flip(bin1, 3));
    cpl_test_error(CPL_ERROR_NONE);

    cpl_test_eq_mask(bin1, bin2);

    cpl_mask_delete(bin2);

    /* Test cpl_mask_move() */
    cpl_test_zero(cpl_mask_move(bin1, 2, new_pos));
    cpl_test_error(CPL_ERROR_NONE);

    /* Test cpl_mask_extract_subsample() */
    bin2 = cpl_mask_extract_subsample(bin1, 2, 2);
    cpl_test_nonnull(bin2);

    cpl_mask_delete(bin1);
    cpl_mask_delete(bin2);

    /* End of tests */
    return cpl_test_end(0);
}



/*----------------------------------------------------------------------------*/
/**
  @brief    Shift a mask using brute force
  @param    in      mask to shift
  @param    x_shift shift to apply in x
  @param    y_shift shift to apply in y
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE
  @see      cpl_image_shift()

 */
/*----------------------------------------------------------------------------*/
static cpl_error_code cpl_mask_shift_bf(cpl_mask * in, 
                                        int        x_shift,
                                        int        y_shift)
{
    int                 xs_abs, ys_abs;
    cpl_mask        *   loc;
    cpl_binary      *   ploc;
    cpl_binary      *   pin;
    int                 i, j;
    const int nx = cpl_mask_get_size_x(in);
    const int ny = cpl_mask_get_size_y(in); 
    /* Initialise */
    xs_abs = abs(x_shift);
    ys_abs = abs(y_shift);
    
   /* Check entries */
    cpl_ensure_code(in != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code((xs_abs<nx)&&(ys_abs<ny), CPL_ERROR_ILLEGAL_INPUT);

    /* Handle the case where nothing has to be done */
    if (x_shift==0 && y_shift==0) return CPL_ERROR_NONE;

    /* Create the local mask */
    loc = cpl_mask_duplicate(in);
    ploc = cpl_mask_get_data(loc);
    pin = cpl_mask_get_data(in);

    for (j=0; j<ny; j++) {
        for (i=0; i<nx; i++) {
            if ((i-x_shift>=0) && (i-x_shift<nx) &&
                    (j-y_shift>=0) && (j-y_shift<ny)) {
                pin[i+j*nx] = ploc[(i-x_shift)+(j-y_shift)*nx];
            } else {
                pin[i+j*nx] = CPL_BINARY_1;
            }
        }
    }
    cpl_mask_delete(loc);
    return CPL_ERROR_NONE;
}


/*----------------------------------------------------------------------------*/
/**
  @brief  Test cpl_mask_shift()
  @param  m      The kernel X-size
  @param  n      The kernel Y-size
  @return void

 */
/*----------------------------------------------------------------------------*/
static void cpl_mask_shift_test(int m, int n)
{

    cpl_image * testimg = cpl_image_new(m, n, CPL_TYPE_FLOAT);
    cpl_mask  * mask0   = cpl_mask_new(m, n);
    cpl_error_code error;
    int i, j;

    cpl_msg_info(cpl_func, "Shift(%d,%d)", m, n);

    error = cpl_mask_shift(NULL, 0, 0);
    cpl_test_error(error);
    cpl_test_eq(error, CPL_ERROR_NULL_INPUT);


    error = cpl_mask_shift(mask0, m, 0);
    cpl_test_error(error);
    cpl_test_eq(error, CPL_ERROR_ILLEGAL_INPUT);

    error = cpl_mask_shift(mask0, -m, 0);
    cpl_test_error(error);
    cpl_test_eq(error, CPL_ERROR_ILLEGAL_INPUT);


    error = cpl_mask_shift(mask0, 0, n);
    cpl_test_error(error);
    cpl_test_eq(error, CPL_ERROR_ILLEGAL_INPUT);

    error = cpl_mask_shift(mask0, 0, -n);
    cpl_test_error(error);
    cpl_test_eq(error, CPL_ERROR_ILLEGAL_INPUT);

    if (m == 1 && n == 1) {
        cpl_mask * mask1 = cpl_mask_new(m, n);
        cpl_test_zero(cpl_mask_shift(mask1, 0, 0));
        cpl_mask_delete(mask1);
    } else {
        cpl_mask * mask1 = cpl_mask_new(m, n);
        cpl_mask * mask2 = cpl_mask_new(m, n);
        int count;

        do {
            /* Need a mask with a mix of elements */
            cpl_test_zero(cpl_image_fill_noise_uniform(testimg, 0.0, 1.0));
            cpl_mask_delete(mask0);
            mask0 = cpl_mask_threshold_image_create(testimg, 0.25, 0.75);
            count = cpl_mask_count(mask0);
        } while (count == 0 || count == m*n);

        for (j = 1-n; j < n; j++) {
            for (i = 1-m; i < m; i++) {
                double now;

                cpl_mask_copy(mask1, mask0, 1, 1);
                cpl_mask_copy(mask2, mask0, 1, 1);

                now = cpl_test_get_cputime();
                cpl_test_zero(cpl_mask_shift(mask1, i, j));
                tshift += cpl_test_get_cputime() - now;

                now = cpl_test_get_cputime();
                cpl_test_zero(cpl_mask_shift_bf(mask2, i, j));
                tshiftbf += cpl_test_get_cputime() - now;

                cpl_test_eq_mask(mask1, mask2);

            }
        }
        cpl_mask_delete(mask1);
        cpl_mask_delete(mask2);
    }

    cpl_image_delete(testimg);
    cpl_mask_delete(mask0);
}
